#!/usr/bin/env python3
"""
This script creates a modelfactors table from GPU hardware counters generated 
by Rocprof. It is a simplified version of the one used by BSC internally.
"""

import argparse
import sys
import os
from pathlib import Path
import pandas as pd
import numpy as np
from typing import List, Tuple
from matplotlib import pyplot as plt
import seaborn as sns
from matplotlib import cm
from matplotlib.colors import Normalize

# Store hardware specifications globally.
DEVICE_SPECS = {
    "MI250X": {
        "Wavefronts": 3520,
    },
}

# Store metrics mapping category to hardware counters globally.
METRICS = {
    "Device Load": [
        "Wavefronts"
    ],
    "Memory Efficiency": [
        "MemUnitBusy", 
        "L2CacheHit", 
        "MemUnitStalled", 
    ],
    "Computational Efficiency": [
        "VALUUtilization", 
        "VALUBusy",
    ],
}


def _parse_user_args() -> argparse.Namespace:
    """Parse user arguments."""
    # Setup argparser.
    parser = argparse.ArgumentParser(
                    prog="gpu_modelfactors.py",
                    description="Create GPU modelfactors from Rocprof " +
                                "GPU hardware counters csv.",
                    )
    parser.add_argument("bin", type=str,
                        help="the binary for which to parse the results.")
    parser.add_argument("-d", "--device", nargs=1, choices=DEVICE_SPECS.keys(),
                        default="MI250X",
                        help="select specific device for percentual coloring.")
    
    # Parse args and just print usage in case no arguments are given.
    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])

    return args


def load_data(bin: str, platform: str) -> List[Tuple[str, pd.DataFrame]]:
    """
    Load profiling results from the provided binary and parse column names.
    """
    dfs = []

    # Check if the results folder exists.
    res_path = f"res-{platform}-{bin}"
    
    if not os.path.isdir(res_path):
        sys.exit(f"Provided 'PLATFORM' and 'bin' created invalid path: {res_path}")
        
    # Loop through results folder and create a df per workload.
    for resdir in Path(res_path).iterdir():
        if not resdir.is_dir():
            continue
        
        resdir_name = resdir.name
        ctr_path = f"{res_path}/{resdir_name}/rocprof/rocprof.csv"
        df = pd.read_csv(ctr_path)
        dfs.append((resdir_name, df))

    return dfs


def prepare_df(dfs: List[pd.DataFrame], device: str) -> pd.DataFrame:
    """Extract only the useful data from the dataframes."""
    # Collect all hardware counters to be filtered in order of plotting.
    hwcs = [counter for v in METRICS.values() for counter in v]
    
    # Setup empty DataFrame to merge data paths.
    merged_df = pd.DataFrame(columns=hwcs)

    for col_name, df in dfs:
        # Extract requested hardware counters.
        new_df = df[hwcs]

        # Set datatypes correctly that are inferred wrongly and add to list.
        new_df.convert_dtypes()
        new_df = new_df.astype({"Wavefronts": int})
        
        # Add first row to merged DataFrame for this input file.
        # Use the provided column name as the index.
        merged_df.loc[col_name] = new_df.iloc[0].values
        
    # Convert Wavefronts to be relative to device specs.
    hwcs_device = [counter for v in [METRICS["Device Load"]] for counter in v]

    for metric in hwcs_device:
        merged_df[metric] = \
            merged_df[metric] / DEVICE_SPECS[device][metric] * 100

    # Return transposed version for the visualization.        
    return merged_df.transpose()


def get_df_percentages(kernel_df: pd.DataFrame, device: str) -> pd.DataFrame:
    """Get percentual changes within the kernel_df and return as DataFrame."""    
    # Create copy of kernel data to store percentages.
    perc_df = kernel_df.copy()

    # For the other metrics, take percentual change compared to the first 
    # column as the baseline.
    hwcs_scaling = [counter 
                   for v in [METRICS["Memory Efficiency"], 
                             METRICS["Computational Efficiency"]] 
                   for counter in v]
    perc_df.loc[hwcs_scaling, perc_df.columns[0]] = 100
    baseline_col = kernel_df.loc[hwcs_scaling, perc_df.columns[0]]

    # Transform other columns to be percentual changes of baseline.
    for col in perc_df.columns[1:]:
        perc_df.loc[hwcs_scaling, col] = 100 + 100 * \
            ((kernel_df.loc[hwcs_scaling, col] - baseline_col) / baseline_col)
    
        # Fix NaNs due to division by zero.
        if perc_df.loc[hwcs_scaling, col].isnull().values.any():
            idxs = np.where(perc_df.loc[hwcs_scaling, col].isnull().values)[0]
            
            for idx in idxs:
                # Check if it the NaN is due to a division by zero.
                # If so, simply add 0% or 100% to the dividend.
                if baseline_col.iloc[idx] == 0.0:
                    # Transform idx to row identifier.
                    metric = hwcs_scaling[idx]
                    perc_df.loc[metric, col] = 100 + kernel_df[col].iloc[idx]

    return perc_df


def plot_modelfactors(kernel_df: pd.DataFrame, device: str, bin: str, platform: str) -> None:
    """Plot the modelfactors for the given kernel as a heatmap."""
    # Transform the DataFrame into one based on percentages for coloring the 
    # cells correctly.
    perc_df = get_df_percentages(kernel_df, device)
    extra_columns = max([0, len(kernel_df.columns) - 4])

    # Create figure with subplots for each category.
    # Add more horizontal space per column added.
    fig = plt.figure(figsize=(6.4 + extra_columns, 4.8))
    hratios = [len(v) for v in METRICS.values()]
    hratios.insert(2, 0)
    gs = fig.add_gridspec(len(METRICS.keys()) + 1, hspace=0.3, 
                          height_ratios=hratios)
    axs = gs.subplots(sharex=True, sharey=False)
    
    # Remove filler axis' styling and remove it from list.
    axs[2].set_axis_off()
    axs = np.delete(axs, 2)

    # Map category to subtext.
    cat_text = {
        "Device Load": "(% of device)",
        "Occupancy": "(% of device)",
        "Memory Efficiency": f"(% of {kernel_df.columns[0]})",
        "Computational Efficiency": f"(% of {kernel_df.columns[0]})",
    }
    
    # Plot the metrics in their categories.
    for ax, [cat, hwcs] in zip(axs, METRICS.items()):
        sns.heatmap(data=perc_df.loc[hwcs], annot=kernel_df.loc[hwcs], 
                    fmt=".2f", cmap="RdYlGn", linewidths=1.5, ax=ax, 
                    cbar=False, vmin=0, vmax=100)

        # Fix label and ticks.
        ax.tick_params(axis="y", labelrotation=0)
        ax.tick_params(axis="x", bottom=False, top=False, labelbottom=False)
        
        # Add category and subtext above first label.
        label_pos = ax.get_yticklabels()[0].get_position()
        ax.text(label_pos[0] - 0.1, label_pos[1] - 0.65, cat, 
                style="oblique", fontweight="heavy", ha="right", 
                fontdict={'family' : 'monospace'})
        ax.text(label_pos[0] - 0.1, label_pos[1] - 0.36, cat_text[cat], 
                style="italic", fontweight="semibold", ha="right", 
                fontsize="xx-small", fontdict={'family' : 'monospace'})

    # Only enable horizontal axis labels at the top.
    axs[0].tick_params(axis="x", top=True, labeltop=True)

    # Add colorbar to the right side of the figure.
    cmap = cm.ScalarMappable(norm=Normalize(0, 100), cmap="RdYlGn")
    fig.colorbar(cmap, label="Relative load (%)", ax=axs)

    # Add title.
    bin = bin.replace("$", "\$")
    fig.suptitle(f"GPU Modelfactors for {bin}.", 
                 x=0.35+extra_columns*0.02, y=0.99, 
                 horizontalalignment="center")
    plt.savefig(f"res-{platform}-{bin}/scaling.png", bbox_inches="tight", 
                dpi=600)


if  __name__ == "__main__":
    # Parse input arguments.
    args = _parse_user_args()

    # Check if PLATFORM is correctly set.
    platform = os.environ.get("PLATFORM")
    options = ["lumi", "mn5"]

    if platform not in options:
        sys.exit(f"Environment variable 'PLATFORM' not valid! Options: {options}.")

    # Load CSV data with hardware counters.
    dfs = load_data(args.bin, platform)

    # Prepare dfs to create a modelfactors table.
    kernel_df = prepare_df(dfs, args.device)

    # Plot the heatmap with comparison.
    plot_modelfactors(kernel_df, args.device, args.bin, platform)
