#!/usr/bin/env bash
# ------------------------------------------------------------------------------
# This file profiles the provided binary on execution for the set platform.
# ------------------------------------------------------------------------------

# The BASEPROF is the PWD of profile.sh, regardless of the callers location.
export BASEPROF=$(cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd)

# Source helpers for colored prints.
source $BASEPROF/../helpers.sh


# Parse the workloads file for the different experiments and store in ARGS.
function parse_workloads {
    ARGS=()
    local exp_id=0
    echo "Parsed workloads:"

    # Read workloads file and parse variables.
    while IFS=" ", read -r nx ny nsteps; do
        # Test for empty variables.
        if [ -z "$nx" ] || [ -z "$ny" ] || [ -z "$nsteps" ]; then
            fatal "workloads.txt contains setup with missing variable:  '[$exp_id] $nx $ny $nsteps'"
            exit $?
        fi

        # Test that all variables consist of only numericals.
        local re="^[0-9]+$"
        if ! [[ $nx =~ $re ]] || ! [[ $ny =~ $re ]] || ! [[ $nsteps =~ $re ]]; then
            fatal "workloads.txt contains a non-numeric:  '[$exp_id] $nx $ny $nsteps'"
            exit $?
        fi

        # Print parsed workload, store in array, and increment counter.
        echo -e "\t[$exp_id] nx: $nx, ny: $ny, nsteps: $nsteps"
        ARGS+=("$nx" "$ny" "$nsteps")
        exp_id=$((exp_id + 1))
    done <<< $(tail -n +3 workloads.txt)
}

# Profile binary on Lumi.
function profile_lumi {
    # Set runtime variables.
    export OMP_NUM_THREADS=1
    export MPICH_GPU_SUPPORT_ENABLED=1

    # Loop over all workloads as set in ARGS by the parse_workloads function.
    for ((i=0; i<${#ARGS[@]}; i=i+3)); do
        nx="${ARGS[$i]}"
        ny="${ARGS[$i+1]}"
        nsteps="${ARGS[$i+2]}"
        echo ""
        info "Running workload:  [$((i/3))] nx: $nx, ny: $ny, nsteps: $nsteps"

        # Create folders for results, or skip iteration if already exist.
        RESULTS="res-$PLATFORM-$1/${nx}_${ny}_${nsteps}"

        if [ -d "$RESULTS" ]; then
            echo -e "\tResults exist, skipping.."
            continue
        fi

        mkdir -p "$RESULTS/rocprof/"

        # Specify where to store the Rocprof hardware counters.
        ROCMFILES="$RESULTS/rocprof/rocprof.csv"

        # Profile the provided binary. First take a trace, then collect HWCs.
        info "[$(date +'%H:%M:%S')]  Profiling '$1'.."
        rocprof --hip-trace --roctx-trace -o $ROCMFILES $BASEPROF/../bin/$1 \
            > $RESULTS/out.log 2> >(tee $RESULTS/err.log >&2)

        rocprof -i $BASEPROF/rocprof_counters.txt -o $ROCMFILES \
            $BASEPROF/../bin/$1 \
            >> $RESULTS/out.log 2> >(tee -a $RESULTS/err.log >&2)

        # Inform according to return value of rocprof.
        retval=$?
        if [ $retval -ne 0 ]; then
            fatal "[$(date +'%H:%M:%S')]  Failed profiling '$1' with error code $retval"
        else
            success "[$(date +'%H:%M:%S')]  Finished profiling '$1'"
        fi
    done
}

# Profile binary on MareNostrum 5.
function profile_mn5 {
    # TODO: Implement.
    echo "To be created.."
}

# Entry function.
function main {
    # Check if platform is set correctly.
    platform_check

    # Check if provided binary exists.
    command -v $BASEPROF/../bin/$1 >/dev/null @>&1 || \
        { fatal "Binary '$1' is not in bin folder."; exit $?; }

    # Parse workloads from the user, which are stored in ARGS variable.
    parse_workloads

    # Load modules.
    source $BASEPROF/../load_modules.sh

    # Profile provided binary on set platform.
    if [ "$PLATFORM" == "lumi" ]; then
        profile_lumi $1
        # sbatch profile_lumi.sh $1
    elif [ "$PLATFORM" == "mn5" ]; then
        profile_mn5 $1
    fi
}

main "$@"
